/*
 * Copyright (c) 2012 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.zeitsync.AuftragsListe
 *
 * de.bsvrz.kex.tls.zeitsync.AuftragsListe is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.kex.tls.zeitsync.AuftragsListe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.kex.tls.zeitsync.AuftragsListe; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.zeitsync;

//~ JDK IMPORTE ===============================================================

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

//~ KLASSEN ===================================================================

/**
 * Singletonklasse, welche die Auftragsliste für die ZeitSynchronisation verwaltet.
 * <p>
 * Die Auftragsliste ermöglicht das Hinzufügen von Aufträgen ({@link Auftrag} sowie die Entnahme des aktuell
 * auszuführenden Auftrags. Die Zugriffe auf die Auftragsliste ist intern synchronisiert.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class AuftragsListe {
    private static AuftragsListe       dieEinzigeInstanz = new AuftragsListe();
    private static final List<Auftrag> _listeAuftraege   = new ArrayList<Auftrag>();

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * Privater Konstruktor.
     */
    private AuftragsListe() {}

    //~ METHODEN ==============================================================

    /**
     * Fügt einen neuen Auftrag zur Liste der Aufträge hinzu. Die Liste wird intern so sortiert, dass die Methode {@link
     * #holeNaechstenAuftrag()} immer den zum aktuellen Zeitpunkt nächsten auszuführenden Auftrag zurückliefert.
     *
     * @param auftrag Neuer Auftrag, der in die Liste der Aufträge eingetragen werden soll.
     */
    public void addAuftrag(Auftrag auftrag) {
        synchronized (_listeAuftraege) {

            // Zuerst bisherigen (zeitlich alten) Auftrag entfernen...
            for (Auftrag auftragAlt : _listeAuftraege) {
                if (auftragAlt.istAelterAls(auftrag)) {
                    _listeAuftraege.remove(auftragAlt);

                    break;
                }
            }

            // .. und dann den zeitlich aktuellen Auftrag eintragen
            _listeAuftraege.add(auftrag);
            Collections.sort(_listeAuftraege, new Comparator<Auftrag>() {
                public int compare(Auftrag auftrag1, Auftrag auftrag2) {

                    // Zeitliche Sortierung der Aufträge nach Ausführungszeitpunkt
                    if (auftrag1.getAusfuehrungsZeitPunkt() < auftrag2.getAusfuehrungsZeitPunkt()) {
                        return -1;
                    }

                    if (auftrag1.getAusfuehrungsZeitPunkt() > auftrag2.getAusfuehrungsZeitPunkt()) {
                        return 1;
                    }

                    return 0;
                }
            });
        }
    }

    //~ GET METHODEN ==========================================================

    /**
     * Liefert die Instanz des Singletons.
     *
     * @return Instanz des Singletons
     */
    public static AuftragsListe getInstanz() {
        return dieEinzigeInstanz;
    }

    //~ METHODEN ==============================================================

    /**
     * Gibt zurück, ob noch Aufträge (aktuell oder zukünftig) anstehen.
     *
     * @return <code>true</code>, wenn noch Aufträge (aktuell oder zukünftig) anstehen, andernfalls <code>false</code>.
     */
    public boolean hatNochAuftraege() {
        synchronized (_listeAuftraege) {
            return _listeAuftraege.size() != 0;
        }
    }

    /**
     * Liefert eine Kopie der Auftragsliste, z.B. für Tests oder Debugausgaben des aktuellen Stands zu einem
     * bestimmten Zeitpunkt.
     *
     * @return Kopie der Auftragsliste.
     */
    public List<Auftrag> holeKopieAuftragsListe() {
        List<Auftrag> listeKopieAuftraege = new ArrayList<Auftrag>();

        synchronized (_listeAuftraege) {
            for (Auftrag auftrag : _listeAuftraege) {
                listeKopieAuftraege.add(auftrag);
            }
        }

        return listeKopieAuftraege;
    }

    /**
     * Liefert den, bezogen auf den aktuellen Zeitpunkt, nächsten auszuführenden Auftrag zurück.
     *
     * @return Nächster auszuführender Auftrag oder <code>null</code>, wenn kein Auftrag aktuell auszuführen ist. Der
     *         zurückgegebene Auftrag wird entfernt und kann nicht noch einmal abgerufen werden.
     */
    public Auftrag holeNaechstenAuftrag() {
        Auftrag naechsterAuftrag = null;

        synchronized (_listeAuftraege) {
            if ((_listeAuftraege.size() > 0) && (_listeAuftraege.get(0).getAusfuehrungsZeitPunkt() < System.currentTimeMillis())) {
                naechsterAuftrag = _listeAuftraege.get(0);
                _listeAuftraege.remove(0);
            }
        }

        return naechsterAuftrag;
    }

    /**
     * Lösche alle Aufträge aus der Liste.
     */
    public void loescheAlleAuftraege() {
        _listeAuftraege.clear();
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 06.07.12
