/*
 * Copyright (c) 2012 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.zeitsync.ZeitSynchronisation
 *
 * de.bsvrz.kex.tls.zeitsync.ZeitSynchronisation is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.kex.tls.zeitsync.ZeitSynchronisation is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.kex.tls.zeitsync.ZeitSynchronisation; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.zeitsync;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

//~ KLASSEN ===================================================================

/**
 * Klasse verwaltet den Prozess für die Zeitsynchronisation.
 * <p>
 * Die zu synchronisierten SteuerModule, sowie DE (0 oder 255) und Intervall
 * sind im DaV-Objekt VewTlsZeitSync, das über den Aufrufparameter übergeben
 * wird, definiert.
 * <p>
 * Prozess prüft in regelmäßigen Abständen (wird über den Aufrufparameter definiert)
 * nach den Zeitsynchronisation-Aufträgen und führt die aktuellen aus.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Liliya Givorgizova (LG)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ZeitSynchronisation implements StandardApplication, ClientReceiverInterface {
    private static final String ASPEKT = "asp.parameterSoll";
    private static final String ATG    = "atg.paramTlsZeitSync";

    /** DebugLogger für Debug-Ausgaben. */
    private static Debug           debug = Debug.getLogger();
    private static DataDescription datenBeschreibung;

    //~ FELDER ================================================================

    /** Datenverteilerverbindung. */
    private ClientDavInterface _dav;

    /** Die Konfiguration des DaV. */
    private DataModel _davKonf;

    /** Parameter: Wartezeit für die Auftragsüberwachung in Stunden. */
    private int _parPauseInStunden;

    /** Parameter: SystemObjekt-PID für die Zeit-Synchronisation. */
    private String _parSoPid;

    /** Wartezeit für die Auftragsüberwachung in Millisekunden. */
    private long _pauseInMillis;

    //~ METHODEN ==============================================================

    /**
     * Fügt einen neuen Auftrag. Ältere Auftrage für eine solche Datenidentifikation werden
     * automatisch ersetzt, so dass je Datenidentifikation maximal ein Auftrag anliegt.
     * Eingefügt werden nur aktivierte Aufträge.
     *
     * @param smPid SteuerModul PID.
     * @param deKanal DE-Kanal (0 oder 255).
     * @param ausfuehrungsZeitPunkt Zeitpunkt der geplanten Ausführung für diesen Auftrag in Millisekunden.
     * @param zeitIntervallInMillis ZeitIntervall in Millis zwischen den Ausführen der Aufträge.
     * @param aktiviert Legt fest, ob der Auftragt aktiviert ist (Wert == true).
     */
    private void addAuftrag(String smPid, int deKanal, long ausfuehrungsZeitPunkt, long zeitIntervallInMillis, boolean aktiviert) {

        // Eingefügt werden nur aktivierte Aufträge:
        if (!aktiviert) {
            return;
        }

        // Auftrag nur erzeugen, wenn zeitIntervallInMillis > 0 ist:
        if (zeitIntervallInMillis > 0) {
            Auftrag auftrag = new Auftrag(_dav, smPid, deKanal, ausfuehrungsZeitPunkt, zeitIntervallInMillis, aktiviert);

            AuftragsListe.getInstanz().addAuftrag(auftrag);

            //
            debug.info(String.format("Eingefügter Auftrag: %s", auftrag));
        }
    }

    @Override
    public void initialize(ClientDavInterface dav) throws Exception {

        // --------------------------------------------------------------
        // DaV-Zugriff:
        _dav     = dav;
        _davKonf = _dav.getDataModel();

        // --------------------------------------------------------------
        // Initialisiere ZeitSync-Objekt:
        SystemObject soZeitSync = _davKonf.getObject(_parSoPid);

        if (soZeitSync == null) {
            debug.error(String.format("Das Objekte mit dem PID <%s> kann nicht ermittelt werden!. Bitte prüfen Sie den Aufrufparameter <-soZeitsyncPid>. Programm wird beendet.", _parSoPid));

            throw new RuntimeException("Null-Objekt");
        }

        // --------------------------------------------------------------
        // Initialisiere DaV-Daten:
        AttributeGroup atg = _davKonf.getAttributeGroup(ATG);
        Aspect         asp = _davKonf.getAspect(ASPEKT);

        datenBeschreibung = new DataDescription(atg, asp);

        // --------------------------------------------------------------
        // DaV-Anmeldung als Empfänger:
        try {
            _dav.subscribeReceiver(this, soZeitSync, datenBeschreibung, new ReceiveOptions(false), ReceiverRole.receiver());
        }
        catch (Exception e) {
            debug.error(String.format("Fehler bei der Anmeldung der Daten als Empfänger: <%s>, <%s>", soZeitSync, datenBeschreibung), e);

            throw new RuntimeException(e);
        }
    }

    /**
     * Main-Methode.
     *
     * @param args Parameter.
     */
    public static void main(String[] args) {

        //
        ZeitSynchronisation zeitSync = new ZeitSynchronisation();

        StandardApplicationRunner.run(zeitSync, args);

        // Debug erst hier initialisieren, da erst ab hier verfügbar !!
        debug = Debug.getLogger();

        // Starte den Überwachungsprozess:
        zeitSync.starteProzess();
    }

    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        String par;

        // --------------------------------------------------------------
        // Hole den Parameter 'SystemObjekt PID für die Zeit-Synchronisation':
        par = "-soZeitsyncPid";

        try {
            _parSoPid = argumentList.fetchArgument(par).asNonEmptyString();
        }
        catch (Exception e) {
            debug.error(String.format("Bitte die SystemObjekt-PID für die Zeit-Synchronisation als Parameter <%s> übergeben!", par), e);

            throw new RuntimeException(e);
        }

        // --------------------------------------------------------------
        // Hole den Parameter 'Pause in Stunden':
        par = "-ueberwachungsZeitInStunden";

        try {
            _parPauseInStunden = argumentList.fetchArgument(par).intValue();
            _pauseInMillis     = _parPauseInStunden * 60 * 60 * 1000;
        }
        catch (Exception e) {
            debug.error(String.format("Bitte die Pause in Stunden für die Zeit-Synchronisation-Überwachung als Parameter <%s> übergeben!", par), e);

            throw new RuntimeException(e);
        }
    }

    /**
     * Starte den Überwachungsprozess:
     * regelmäßig prüfe die Auftragsliste und führe aktuelle Aufträge aus.
     */
    private void starteProzess() {

        // --------------------------------------------------------------
        // debug-Info:
        debug.info(String.format("%n=================================================================%n" + "ZeitSynchronisation gestartet...%n" + "=================================================================%n"));

        // --------------------------------------------------------------
        // Endlosschleife mit der Prüfung der Auftragsliste:
        while (true) {
            Auftrag auftrag;

            // --------------------------------------------------------------
            // Falls es einen Auftrag zum Ausführen in der Liste gibt:
            if ((auftrag = AuftragsListe.getInstanz().holeNaechstenAuftrag()) != null) {

                // -------------------------------------
                // Füge neuen Auftrag mit gleicher Datenidentifikation und nächstem Ausführungspunkt:
                long ausfuehrungsZeitPunkt = System.currentTimeMillis() + auftrag.getZeitIntervallInMillis();

                addAuftrag(auftrag.getSmPid(), auftrag.getDeKanal(), ausfuehrungsZeitPunkt, auftrag.getZeitIntervallInMillis(), auftrag.isAktiviert());

                // -------------------------------------
                // Ausführen den aktuellen Auftrag:
                if (auftrag.ausfuehren()) {
                    debug.info(String.format("Ausgeführter Auftrag: %s", auftrag));
                } else {
                    debug.warning(String.format("Folgender Auftrag wurde nicht ausgeführt: %s%n(siehe Meldungen oben)", auftrag));
                }
            }

            // --------------------------------------------------------------
            // Zurzeit ist nichts zu tun, also warten und es dann noch mal versuchen...
            else {
                debug.fine(String.format(                                                                     //
                    "Zur Zeit keine sofort auszuführenden Aufträge in der Auftragsliste. Warte %d Stunden.",  //
                    _parPauseInStunden));

                // Pause
                try {
                    Thread.sleep(_pauseInMillis);
                }
                catch (InterruptedException e) {
                    debug.info("Sleep der Auftragsbearbeitung der ZeitSynchronisation wurde vorzeitig unterbrochen: ", e);
                }
            }
        }
    }

    @Override
    public void update(ResultData[] results) {
        debug.info(String.format("%n=================================================================%n" + "Erstelle die Liste der Auftrage aus DaV-Daten neu...%n" + "=================================================================%n"));

        // --------------------------------------------------------------
        // Lösche alle Aufträge:
        AuftragsListe.getInstanz().loescheAlleAuftraege();

        // --------------------------------------------------------------
        // Erstelle die Liste der Auftrage aus DaV-Daten neu:
        for (ResultData daten : results) {
            if (daten.hasData()) {
                String  smPid;
                int     deKanal;
                long    ausfuehrungsZeitPunkt = System.currentTimeMillis();
                long    zeitIntervallInMillis;
                boolean aktiviert;

                //
                Data.Array ar = daten.getData().getArray("SmZeitSyncParameter");
                Data       data;

                for (int i = 0; i < ar.getLength(); i++) {
                    data                  = ar.getItem(i);
                    smPid                 = data.getReferenceValue("SteuerModulReferenz").getSystemObjectPid();
                    deKanal               = data.getUnscaledValue("De0Oder255").intValue();
                    zeitIntervallInMillis = data.getTimeValue("Intervall").getMillis();
                    aktiviert             = (data.getTextValue("aktiviert").getText().equals("Ja"));
                    addAuftrag(smPid, deKanal, ausfuehrungsZeitPunkt, zeitIntervallInMillis, aktiviert);
                }
            }
        }

        debug.info(String.format("%n=================================================================%nMomentan sind %d Aufträge in der Liste erstellt.%n=================================================================%n",  //
                                 AuftragsListe.getInstanz().holeKopieAuftragsListe().size()));
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 17.07.12
